# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from ._base import Deduplicator as Deduplicator

from ._monotonic import MonotonicDeduplicator as MonotonicDeduplicator

from ._cyclic import CyclicDeduplicator as CyclicDeduplicator
